#include <xc.h>
#include <stdio.h>

/* Configuration Settings */
// OSC Input = Internal with PLL
// CPU Clock = ( ( 16MHz * 3 ) / 3
// USB Clock is 48MHz
// Disable WDT , Enable PWRT , Enable BOR
// Enable Stack Overflow Reset , Disable MCLR Function , Disable LVP Function
#pragma config FOSC = INTOSC , PLLMULT = 3x , PLLEN = ENABLED , CPUDIV = CLKDIV3
#pragma config USBLSCLK = 48MHz
#pragma config WDTE = OFF , PWRTE = ON , BOREN = ON , BORV = LO
#pragma config STVREN = ON , MCLRE = ON , LVP = OFF

#define SystemFrequency		48000000L / 3 / 4

void InitializeTimer2(void);
void InitializeUSART(void);
void putcUSART(char c);
unsigned char getcUSART(void);
void InitializeI2C(void);
char IdleI2C(void);
char StartI2C(void);
char StopI2C(void);
char ReStartI2C(void);
char NAckI2C(void);
char AckI2C(void);
char WriteI2C(unsigned char Data);
char ReadI2C(unsigned char * Data);

char I2C_1ByteAddrByteWirte(unsigned char SalveAddr, unsigned char Addr, unsigned char Data);
char I2C_1ByteAddrRandomRead(unsigned char SalveAddr, unsigned char Addr, unsigned char * Data);
char I2C_2BytesAddrByteWirte(unsigned char SalveAddr, unsigned int Addr, unsigned char Data);
char I2C_2BytesAddrRandomRead(unsigned char SalveAddr, unsigned int Addr, unsigned char * Data);

int I2C_MCP9800Set12BitsRresolution(unsigned char SalveAddr);
int I2C_MCP9800ReadTemperature(unsigned char SalveAddr, unsigned int * Temperature);

void UART_PutRAMString(unsigned char * String);

#define Timer2Tick		( SystemFrequency / 16 / 64 / Timer2TogglesPerSec )
#define Timer2TogglesPerSec     100

unsigned char UARTString[ 100 ];
unsigned long i;
unsigned char TimeFlag;

enum
{
    getTemperature = 1,
    getMACAddress,
    getTime,
    ToggleLED
} ActivityFlag;


unsigned char Counter = 0;

struct DayTime
{
    unsigned char Year;
    unsigned char Month;
    unsigned char Day;
    unsigned char Hour;
    unsigned char Min;
    unsigned char Sec;
    unsigned char Date;
} NowDayTime;

void main(void)
{
    unsigned char Temporary = 0;

    OSCCONbits.IRCF = 0x0f; // 16 MHz or 48 MHz HF
    OSCCONbits.SCS = 0x00; // Clock determined by FOSC<2:0> in Configuration Words.

    ACTCONbits.ACTSRC = 1; // The HFINTOSC oscillator is tuned using Fll-speed USB events
    ACTCONbits.ACTEN = 1; // ACT is enabled, updates to OSCTUNE are exclusive to the ACT

    LATAbits.LATA5 = 0; // LED1 (D6)
    TRISAbits.TRISA5 = 0;

    ANSELAbits.ANSA4 = 0; // LED2 (D7)
    LATAbits.LATA4 = 0;
    TRISAbits.TRISA4 = 0;

    ANSELCbits.ANSC6 = 0; // Wakeup Hardware
    LATCbits.LATC6 = 1;
    TRISCbits.TRISC6 = 0;

    ANSELCbits.ANSC7 = 0; // Wakeup Software
    LATCbits.LATC7 = 1;
    TRISCbits.TRISC7 = 0;

    ANSELCbits.ANSC2 = 0; // CMD/MLDP
    LATCbits.LATC2 = 0;
    TRISCbits.TRISC2 = 0;

    InitializeTimer2();
    InitializeUSART();
    InitializeI2C();

    // MCP9800 Setup
    I2C_MCP9800Set12BitsRresolution(0x9a);

    //MCP79412 Setup
    // 1.Enable Crystail
    I2C_1ByteAddrRandomRead(0xDE, 0x00, &Temporary);
    if ((Temporary & 0x80) != 0x80)
        I2C_1ByteAddrByteWirte(0xDE, 0x00, 0x80);
    for (i = 0; i < 2000; i++);

    // 2.Enable VBAT
    I2C_1ByteAddrRandomRead(0xDE, 0x03, &Temporary);
    if ((Temporary & 0x08) != 0x08)
        I2C_1ByteAddrByteWirte(0xDE, 0x03, 0x08);
    for (i = 0; i < 2000; i++);

    LATCbits.LATC2 = 0; // CMD Mode
    sprintf(UARTString, "sf,2\r\n");
    UART_PutRAMString(UARTString);
    sprintf(UARTString, "sr,30000800\r\n");
    UART_PutRAMString(UARTString);
    sprintf(UARTString, "r,1\r\n");
    UART_PutRAMString(UARTString);
    LATCbits.LATC2 = 1; // MLDP Mode

    for (i = 0; i < 2000; i++);

    INTCONbits.PEIE = 1;
    INTCONbits.GIE = 1;

    while (1)
    {
        unsigned int i = 0;
        unsigned char MACAddress[8];
        unsigned int Temperature = 0, Integer = 0, Fractional = 0;

        if (TimeFlag)
        {
            TimeFlag = 0;

            switch (ActivityFlag)
            {
            default:
            case getTemperature:
                I2C_MCP9800ReadTemperature(0x9a, &Temperature);
                Integer = Temperature >> 8;
                Fractional = (Temperature & 0x00f0) >> 4;
                Fractional *= 625;

                sprintf(UARTString, "[%03d]Temperature:\r\n%d.%04d\'C\r\n",
                        ++Counter, Integer, Fractional);
                UART_PutRAMString(UARTString);
                break;

            case getTime:
                // MCP79411
                // Update Date & Time
                I2C_1ByteAddrRandomRead(0xDE, 0x00, &Temporary); // Sec
                NowDayTime.Sec = Temporary & 0x7f;
                for (i = 0; i < 1000; i++);
                I2C_1ByteAddrRandomRead(0xDE, 0x01, &Temporary); // Min
                NowDayTime.Min = Temporary & 0x7f;
                for (i = 0; i < 1000; i++);
                I2C_1ByteAddrRandomRead(0xDE, 0x02, &Temporary); // Hour
                NowDayTime.Hour = Temporary & 0x3f;
                for (i = 0; i < 1000; i++);
                I2C_1ByteAddrRandomRead(0xDE, 0x04, &Temporary); // Date
                NowDayTime.Date = Temporary & 0x3f;
                for (i = 0; i < 1000; i++);
                I2C_1ByteAddrRandomRead(0xDE, 0x05, &Temporary); // Month
                NowDayTime.Month = Temporary & 0x1f;
                for (i = 0; i < 1000; i++);
                I2C_1ByteAddrRandomRead(0xDE, 0x06, &Temporary); // Year
                NowDayTime.Year = Temporary & 0x3f;
                for (i = 0; i < 1000; i++);
                I2C_1ByteAddrRandomRead(0xDE, 0x03, &Temporary); // Day
                NowDayTime.Day = Temporary & 0x3f;
                for (i = 0; i < 1000; i++);

                sprintf(UARTString, "Time:\r\n%d%d:%d%d:%d%d\r\n",
                        (NowDayTime.Hour >> 4) & 0x0f,
                        NowDayTime.Hour & 0x0f,
                        (NowDayTime.Min >> 4) & 0x0f,
                        NowDayTime.Min & 0x0f,
                        (NowDayTime.Sec >> 4) & 0x0f,
                        NowDayTime.Sec & 0x0f);

                UART_PutRAMString(UARTString);
                break;

            case getMACAddress:
                for (i = 0; i < 8; i++)
                    I2C_1ByteAddrRandomRead(0xCE, 0xf0 + i, &MACAddress + i);

                sprintf(UARTString, "MAC Address:\r\n%02X-%02X-%02X-%02X-%02X-%02X-%02X-%02X\r\n",
                        MACAddress[0], MACAddress[1], MACAddress[2],
                        MACAddress[3], MACAddress[4], MACAddress[5],
                        MACAddress[6], MACAddress[7]);
                
                UART_PutRAMString(UARTString);
                break;
            }
        }
    }

    while (1);
}

void interrupt ISR(void)
{
    unsigned char ReceivedData = 0;

    if (PIR1bits.RCIF && PIE1bits.RCIE)
    {
        ReceivedData = getcUSART();

        switch (ReceivedData)
        {
        case '0':
            ActivityFlag = getTemperature;
            break;
            
        case '1':
            ActivityFlag = getMACAddress;
            break;

        case '2':
            ActivityFlag = getTime;
            break;

        case '3':
            LATAbits.LATA4 = !LATAbits.LATA4;
            break;
        }
    }

    if (PIR1bits.TMR2IF && PIE1bits.TMR2IE)
    {
        static unsigned int Tick = 0;

        PIR1bits.TMR2IF = 0;

        if (++Tick > 100)
        {
            Tick = 0;
            TimeFlag = 1;
            LATAbits.LATA5 = !LATAbits.LATA5;
        }
    }
}

void InitializeTimer2(void)
{
    T2CONbits.T2OUTPS = 0x0f; // 1:16 Postscaler
    T2CONbits.T2CKPS = 0x03; // 1:64 Prescaler
    PR2 = Timer2Tick;
    T2CONbits.TMR2ON = 1;
    PIR1bits.TMR2IF = 0;
    PIE1bits.TMR2IE = 1;
}

void UART_PutRAMString(unsigned char * String)
{
    while (*String != 0x00)
    {
        while (!PIR1bits.TXIF);
        PIR1bits.TXIF = 0;
        putcUSART(*String++);
    }
}

void InitializeUSART(void)
{
    unsigned char c;

    TXSTAbits.TXEN = 1; // TX Enable
    TXSTAbits.TX9 = 0; // 8 Bits Bata
    TXSTAbits.SYNC = 0; // Asynchronous

    RCSTAbits.CREN = 1; // Enables receiver

    TXSTAbits.BRGH = 1; // High Baud Rate
    BAUDCONbits.BRG16 = 1; // 16-bit Baud Rate Generator
    SPBRGL = 34; // 34 for 115200, Base 4MHz Fcy
    SPBRGH = 00;

    ANSELBbits.ANSB5 = 0; // Rx

    RCSTAbits.SPEN = 1; // Serial Port Enable

    PIR1bits.RCIF = 0;
    PIE1bits.RCIE = 1;

    c = RCREG; // read
}

void putcUSART(char c)
{
    TXREG = c;
}

unsigned char getcUSART(void)
{
    char c;

    if (RCSTAbits.OERR) // in case of overrun error
    { // we should never see an overrun error, but if we do,
        RCSTAbits.CREN = 0; // reset the port
        c = RCREG;
        RCSTAbits.CREN = 1; // and keep going.
    }
    else
    {
        c = RCREG;
    }

    return c;
}

void InitializeI2C(void)
{
    SSPCON1bits.SSPM = 0x08; // I2C Master mode, clock = FOSC / (4 * (SSPADD+1))
    SSPSTATbits.SMP = 0; // Slew rate for High-Speed mode (400 kHz)
    SSPSTATbits.CKE = 0; // Disable SMbus specific inputs

    TRISBbits.TRISB4 = 1; // SDA
    ANSELBbits.ANSB4 = 0;
    TRISBbits.TRISB6 = 1; // SCL

    SSPADD = 0x27;

    SSPCON1bits.SSPEN = 1; // Enables
}

char IdleI2C(void)
{
    unsigned long Timeout;

    Timeout = 10;
    // Idle;
    while (SSPCON2bits.ACKEN || SSPCON2bits.RCEN || SSPCON2bits.PEN || SSPCON2bits.RSEN || SEN || SSPSTATbits.R_nW)
    {
        if (Timeout-- == 0)
            return (-1);
    }
    return (0);
}

char StartI2C(void)
{
    unsigned long Timeout;

    Timeout = 100;
    // Start Condition
    SSPCON2bits.SEN = 1;
    while (SSPCON2bits.SEN)
    {
        if (Timeout-- == 0)
            return (-1);
    }
    return (0);
}

char StopI2C(void)
{
    unsigned long Timeout;

    Timeout = 100;
    // Stop Condition
    SSPCON2bits.PEN = 1;
    while (SSPCON2bits.PEN)
    {
        if (Timeout-- == 0)
            return (-1);
    }
    return (0);
}

char ReStartI2C(void)
{
    unsigned long Timeout;

    Timeout = 100;
    // ReStart Condition
    SSPCON2bits.RSEN = 1;
    while (SSPCON2bits.RSEN)
    {
        if (Timeout-- == 0)
            return (-1);
    }
    return (0);
}

char NAckI2C(void)
{
    unsigned long Timeout;

    Timeout = 100;
    // NACK Condition
    SSPCON2bits.ACKDT = 1;
    SSPCON2bits.ACKEN = 1;
    while (SSPCON2bits.ACKEN)
    {
        if (Timeout-- == 0)
            return (-1);
    }
    return (0);
}

char AckI2C(void)
{
    unsigned long Timeout;

    Timeout = 100;
    // NACK Condition
    SSPCON2bits.ACKDT = 0;
    SSPCON2bits.ACKEN = 1;
    while (SSPCON2bits.ACKEN)
    {
        if (Timeout-- == 0)
            return (-1);
    }
    return (0);
}

char WriteI2C(unsigned char Data)
{
    unsigned long Timeout;

    SSPBUF = Data;

    if (SSPCON1bits.WCOL)
    {
        return (-2);
    }
    else
    {
        Timeout = 100;
        while (SSPSTATbits.BF)
        {
            if (Timeout-- == 0)
                return (-1);
        }
        if (IdleI2C())
            return (-3);
        if (SSPCON2bits.ACKSTAT)
            return (-4);

        else return (0);
    }
}

char ReadI2C(unsigned char * Data)
{
    unsigned long Timeout;

    Timeout = 100;
    SSP1CON2bits.RCEN = 1;
    while (SSPCON2bits.RCEN)
    {
        if (Timeout-- == 0)
            return (-1);
    }

    Timeout = 10;
    while (!SSP1STATbits.BF)
    {

        if (Timeout-- == 0)
            return (-1);
    }

    * Data = SSP1BUF;

    return (0);
}

char I2C_1ByteAddrByteWirte(unsigned char SalveAddr, unsigned char Addr, unsigned char Data)
{
    if (IdleI2C())
        return (-1);

    if (StartI2C())
        return (-1);

    if (WriteI2C(SalveAddr & 0xfe))
        return (-1);

    if (WriteI2C(Addr & 0xff))
        return (-1);

    if (WriteI2C(Data))
        return (-1);

    if (StopI2C())
        return (-1);

    return (0);
}

char I2C_1ByteAddrRandomRead(unsigned char SalveAddr, unsigned char Addr, unsigned char * Data)
{
    if (IdleI2C())
        return (-1);

    if (StartI2C())
        return (-1);

    if (WriteI2C(SalveAddr & 0xfe))
        return (-1);

    if (WriteI2C(Addr & 0xff))
        return (-1);

    if (ReStartI2C())
        return (-1);

    if (WriteI2C(SalveAddr | 0x01))
        return (-1);

    if (ReadI2C(Data))
        return (-1);

    if (NAckI2C())
        return (-1);

    if (StopI2C())
        return (-1);

    return (0);
}

char I2C_2BytesAddrByteWirte(unsigned char SalveAddr, unsigned int Addr, unsigned char Data)
{
    if (IdleI2C())
        return (-1);

    if (StartI2C())
        return (-1);

    if (WriteI2C(SalveAddr & 0xfe))
        return (-1);

    if (WriteI2C((Addr >> 8) & 0xff))
        return (-1);

    if (WriteI2C(Addr & 0xff))
        return (-1);

    if (WriteI2C(Data))
        return (-1);

    if (StopI2C())
        return (-1);

    return (0);
}

char I2C_2BytesAddrRandomRead(unsigned char SalveAddr, unsigned int Addr, unsigned char * Data)
{
    if (IdleI2C())
        return (-1);

    if (StartI2C())
        return (-1);

    if (WriteI2C(SalveAddr & 0xfe))
        return (-1);

    if (WriteI2C((Addr >> 8) & 0xff))
        return (-1);

    if (WriteI2C(Addr & 0xff))
        return (-1);

    if (ReStartI2C())
        return (-1);

    if (WriteI2C(SalveAddr | 0x01))
        return (-1);

    if (ReadI2C(Data))
        return (-1);

    if (NAckI2C())
        return (-1);

    if (StopI2C())
        return (-1);

    return (0);
}

int I2C_MCP9800Set12BitsRresolution(unsigned char SalveAddr)
{
    if (IdleI2C())
        return (-1);

    if (StartI2C())
        return (-1);

    if (WriteI2C(SalveAddr & 0xfe))
        return (-1);

    if (WriteI2C(0x01)) // Sensor Configuration Register (CONFIG)
        return (-1);

    if (WriteI2C(0x60))
        return (-1);

    if (StopI2C())
        return (-1);

    return (0);
}

int I2C_MCP9800ReadTemperature(unsigned char SalveAddr, unsigned int * Temperature)
{
    unsigned int Temporary = 0;

    if (IdleI2C())
        return (-1);

    if (StartI2C())
        return (-1);

    if (WriteI2C(SalveAddr & 0xfe))
        return (-1);

    if (WriteI2C(0x00)) // Ambient Temperature Register (TA)
        return (-1);

    if (ReStartI2C())
        return (-1);

    if (WriteI2C(SalveAddr | 0x01))
        return (-1);

    if (ReadI2C(&Temporary))
        return (-1);

    * Temperature = Temporary << 8;

    if (AckI2C())
        return (-1);

    if (ReadI2C(&Temporary))
        return (-1);

    * Temperature += Temporary;

    if (NAckI2C())
        return (-1);

    if (StopI2C())
        return (-1);

    return (0);
}